from kotoba.kotoba import Kotoba
from imagination.decorator.validator import restrict_type
from imagination.meta.package        import Parameter

class ParameterParsingMixin(object):
    @restrict_type(Kotoba)
    def _get_params(self, node):
        package = Parameter()

        index = 0

        for param in node.children('param'):
            try:
                assert param.attribute('name')\
                    and param.attribute('type'),\
                    'The parameter #{} does not have either name or type.'.format(index)
            except AssertionError as e:
                raise IncompatibleBlockError(e.message)

            index += 1
            name   = param.attribute('name')

            if name in package.kwargs:
                raise DuplicateKeyWarning('There is a parameter name "{}" with that name already registered.'.format(name))
                pass

            package.kwargs[name] = self._transformer.cast(
                param,
                param.attribute('type')
            )

        return package



'''
:Author: Juti Noppornpitak

The module contains the entity locator used to promote reusability of components.

.. note::
    Copyright (c) 2012 Juti Noppornpitak

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is furnished to do
    so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
    INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
    PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
    OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

'''
from kotoba import load_from_file

from imagination.factorization import Factorization
from imagination.entity        import CallbackProxy
from imagination.entity        import Entity
from imagination.exception     import *
from imagination.proxy         import Proxy

class Locator(object):
    ''' Entity locator '''

    __self_reference__ = 'container'

    def __init__(self):
        self._entities          = {}
        self._tag_to_entity_ids = {}
        self._in_passive_mode   = False

    @property
    def in_passive_mode(self):
        return self._in_passive_mode

    @in_passive_mode.setter
    def in_passive_mode(self, value):
        if not isinstance(value, bool):
            raise ValueError('Expected a boolean value for in_passive_mode.')

        self._in_passive_mode = value

    def fork(self, id):
        '''
        Fork the entity identified by ``id``.

        :param `id`: entity identifier
        '''
        if id == self.__self_reference__:
            return self

        try:
            entity = self.get_wrapper(id)

            # Prevent the locator from forking a new instance of Proxy or CallbackProxy.
            if isinstance(entity, CallbackProxy) or isinstance(entity, Proxy):
                raise ForbiddenForkError('Unable to fork {}'.format(type(entity).__name__))

            return entity.fork()
        except KeyError:
            raise UnknownEntityError('The requested entity named "{}" is not found.'.format(id))

    @property
    def entity_identifiers(self):
        keys = list(self._entities.keys())
        keys.append(self.__self_reference__)
        keys.sort()

        return keys

    def get(self, id):
        '''
        Retrieve an instance of the entity

        :param `id`: entity identifier
        :returns: an instance of the requested entity
        '''
        if id == self.__self_reference__:
            return self

        try:
            entity = self.get_wrapper(id)

            # Retrieve a callback entity (callable).
            if isinstance(entity, CallbackProxy):
                #print('GET {} "{}" [Done]'.format(type(entity).__name__, id))
                return entity()

            # Retrieve a proxy object to an entity
            if isinstance(entity, Proxy):
                #print('GET {} "{}" [Done]'.format(type(entity).__name__, id))
                return entity #return the proxy reference.

            if isinstance(entity, Factorization):
                #print('GET {} "{}" [Done]'.format(type(entity).__name__, id))
                return entity.fork()

            #print('GET {} "{}" [Done]'.format(type(entity).__name__, id))

            return entity.instance if isinstance(entity, Entity) else entity
        except UnknownEntityError as e:
            raise UnknownEntityError(
                'The requested entity named "{id}" is not found. This locator only knows {known_keys}. (Reason: {original})'.format(
                    id         = id,
                    known_keys = ', '.join(self._entities.keys()),
                    original   = e
                )
            )

    def get_wrapper(self, id):
        '''
        Retrieve the entity wrapper identified by ``id``.

        :param `id`: entity identifier
        :returns: the requested entity wrapper
        '''
        if id == self.__self_reference__:
            return self

        try:
            return self._entities[id]
        except KeyError:
            if self.in_passive_mode:
                return Proxy(self, id)

            raise UnknownEntityError('The requested entity named "{}" is not found.'.format(id))

    def find_by_tag(self, tag_label):
        '''
        Retrieve entities by *tag_label*.

        :param `tag_label`: tag label
        '''

        # First, get the entity identifiers.
        if tag_label not in self._tag_to_entity_ids:
            return []

        # Then, get references to entities.
        return [self.get(entity_id) for entity_id in self._tag_to_entity_ids[tag_label]]

    def set(self, id, entity):
        ''' Set the given *entity* by *id*. '''

        is_proxy    = isinstance(entity, Proxy)
        is_callback = isinstance(entity, CallbackProxy)

        if type(entity) not in [Entity, Proxy, CallbackProxy, Factorization]:
            raise UnknownEntityError('The type of the given entity named "{}" is not excepted. ({})'.format(id, type(entity).__name__))

        if isinstance(entity, Entity):
            entity.lock()

        if is_proxy and id in self._entities:
            #print('SET Proxy "{}" [Ignored]'.format(id))
            return

        self._entities[id] = entity

        if is_proxy:
            #print('SET Proxy "{}" [Done]'.format(id))
            return

        if is_callback:
            #print('SET Callback "{}" [Done]'.format(id))
            return

        for tag in entity.tags:
            if tag not in self._tag_to_entity_ids:
                self._tag_to_entity_ids[tag] = []

            self._tag_to_entity_ids[tag].append(entity.id)

        #print('SET {} "{}" [Done]'.format(type(entity).__name__, id))

    def has(self, id):
        ''' Check if the entity with *id* is already registered. '''
        return id in self._entities

    def load_xml(self, file_path):
        '''
        Load the entities from a XML configuration file at *file_path*.

        :Status: Deprecated in 1.5

        '''
        raise DeprecatedAPI('Use imagination.helper.assembler.Assembler.load instead.')



'''
:Author: Juti Noppornpitak
:Availability: 1.9

The module contains the package entity used to be an intermediate between :class:`imagination.locator.Locator`
and :class:`imagination.loader.Loader` and simulate the singleton class on the package in the Loader.

.. note::
    Copyright (c) 2015 Juti Noppornpitak

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is furnished to do
    so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
    INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
    PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
    OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

'''

from imagination.decorator.validator import restrict_type
from imagination.common              import InterceptableObject
from imagination.proxy               import Proxy

class NonCallableFactoryMethod(RuntimeError):
    """ The factory method is not callable. """

class NotReadyError(RuntimeError):
    """ When the factory service is not ready to use. """

class Factorization(InterceptableObject):
    """
    Factorization Entity

    :param imagination.locator.Locator locator:    the locator
    :param str factory_id:     the factory entity identifier
    :param str factory_method: the factory method

    .. versionadded:: 1.9
    """
    def __init__(self, locator, factory_id, factory_method, parameters):
        super(Factorization, self).__init__()

        self._locator        = locator
        self._factory_id     = factory_id
        self._factory_method = factory_method
        self._parameters     = parameters
        self._reference      = None
        self._tags     = []

    @property
    def id(self):
        ''' Get the identifier of the proxy. '''
        return self.__id

    @property
    def tags(self):
        '''
        Retrieve the entity tags.

        :rtype: list
        '''
        return self._tags

    @tags.setter
    @restrict_type(list)
    def tags(self, tags):
        '''
        Define the entire entity tags.

        :param tags: new tags as replacements
        :type  tags: list or tuple
        '''
        if self.locked:
            raise LockedEntityException

        self._tags = tags

    def fork(self):
        ''' Fork the entity. '''
        self._prepare()

        factory = self._locator.get(self._factory_id)

        if isinstance(factory, Proxy):
            raise NotReadyError(self._factory_id)

        factory_method = factory.__getattribute__(self._factory_method)

        if not factory_method or not callable(factory_method):
            raise NonCallableFactoryMethod('{}.{} not available ({})'.format(self._factory_id, self._factory_method, type(factory_method)))

        self._reference = factory_method(*self._parameters.largs, **self._parameters.kwargs)

        if self._interceptable and isinstance(self._reference, object):
            self._bind_interceptions(self._reference, self._interceptions)

        return self._reference

    def _prepare(self):
        p = self._parameters

        for i in range(len(p.largs)):
            if isinstance(p.largs[i], Proxy):
                p.largs[i] = p.largs[i].load()

        for i in p.kwargs:
            if isinstance(p.kwargs[i], Proxy):
                p.kwargs[i] = p.kwargs[i].load()



'''
:Author: Juti Noppornpitak
:Availability: 1.5

The module contains the package entity used to be an intermediate between :class:`imagination.locator.Locator`
and :class:`imagination.loader.Loader` and simulate the singleton class on the package in the Loader.

.. note::
    Copyright (c) 2012 Juti Noppornpitak

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is furnished to do
    so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
    INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
    PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
    OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

'''

class Proxy(object):
    '''
    Proxy to a particular entity known by a certain locator.

    :param `locator`: the locator
    :param `id`: entity identifier
    '''
    def __init__(self, locator, id):
        self.__dict__ = {
            'id':      id,
            'locator': locator,
            'cache':   None,
        }

    def load(self):
        if not self.__dict__['cache']:
            self.__dict__['cache'] = self.__dict__['locator'].get(self.__dict__['id'])

        return self.__dict__['cache']

    def __getattr__(self, name):
        if name in self.__dict__:
            return self.__dict__[name]

        actual_object = self.load()

        if not hasattr(actual_object, name):
            raise AttributeError('{} has no attribute "{}".'.format(type(actual_object).__name__), name)

        return getattr(actual_object, name)



'''
:Author: Juti Noppornpitak

The module contains the package entity used to be an intermediate between :class:`imagination.locator.Locator`
and :class:`imagination.loader.Loader` and simulate the singleton class on the package in the Loader.

.. note::
    Copyright (c) 2012 Juti Noppornpitak

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is furnished to do
    so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
    INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
    PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
    OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

'''
import inspect

from imagination.common              import InterceptableObject
from imagination.decorator.validator import restrict_type
from imagination.exception           import InstantiationError
from imagination.loader              import Loader
from imagination.proxy               import Proxy

class ReferenceProxy(object):
    """ Reference Proxy

    .. codeauthor:: Juti Noppornpitak <juti_n@yahoo.co.jp>
    .. versionadded:: 1.20

    .. warning:: experimental feature
    """
    def __init__(self, reference):
        self.__reference = reference

    @property
    def reference(self):
        return self.__reference

class CallbackProxy(object):
    """ Callback Proxy

        A proxy to a callback function where it executes the method with
        pre-defined parameters whenever it is called. The end result will
        be cached by the proxy.

        .. codeauthor:: Juti Noppornpitak <juti_n@yahoo.co.jp>
        .. versionadded:: 1.6
    """
    def __init__(self, callback, args = [], kwargs = {}, static = False):
        if not callable(callback):
            raise ValueError('The callback object is required for {}.'.format(self.__class__.__name__))

        self.__callback = callback
        self.__args     = args
        self.__kwargs   = kwargs
        self.__static   = static
        self.__executed = False
        self.__result   = None

    def __execute(self):
        return self.__callback(*self.__args, **self.__kwargs)

    def __call__(self):
        if not self.__static:
            return self.__execute()

        if not self.__executed:
            self.__executed = True
            self.__result   = self.__execute()

        return self.__result

class Entity(InterceptableObject):
    '''
    Entity represents the package, reference and instance of the reference.

    :param `id`:     the service identifier (string).
    :param `loader`: a service loader which is an instance of :class:`imagination.loader.Loader`.
    :param `args`:   constructor's parameters
    :type args:      list or tuple
    :param `kwargs`: constructor's parameters
    :type args:      dict

    If the loader is not an instance of :class:`imagination.loader.Loader` or
    any classes based on :class:`imagination.loader.Loader`, the exception
    :class:`imagination.exception.UnknownLoaderError` will be thrown.

    .. note::
        This class is to similar to the decorator `tori.decorator.common.singleton`
        and `tori.decorator.common.singleton_with` from Tori Framework, except
        that it is not a singleton class and so any arbitrary class referenced
        by the *loader* only lasts as long as the entity lives.

    .. note::
        In version 1.5, the entity has the ability to fork an non-supervised
        instance of the reference.
    '''

    @restrict_type(None, Loader)
    def __init__(self, id, loader, *args, **kwargs):
        super(Entity, self).__init__()

        self._id       = id
        self._loader   = loader
        self._args     = args
        self._kwargs   = kwargs
        self._instance = None
        self._tags     = []
        self._prepared = False

    @property
    def id(self):
        '''
        Entity ID

        :rtype: strint or unicode or integer
        '''
        return self._id

    @property
    def loader(self):
        '''
        Package loader

        :rtype: imagination.loader.Loader
        '''
        return self._loader

    @property
    def argument_list(self):
        ''' Get the argument list. '''
        return self._args

    @property
    def argument_dictionary(self):
        ''' Get the argument dictionary. '''
        return self._kwargs

    @property
    def activated(self):
        '''
        Check if the entity is already activated.

        This will also inspects if the entity already loads a singleton instance into the memory.
        '''
        return self._instance is not None

    @property
    def tags(self):
        '''
        Retrieve the entity tags.

        :rtype: list
        '''
        return self._tags

    @tags.setter
    @restrict_type(list)
    def tags(self, tags):
        '''
        Define the entire entity tags.

        :param tags: new tags as replacements
        :type  tags: list or tuple
        '''
        if self.locked:
            raise LockedEntityException

        self._tags = tags

    @property
    def instance(self):
        ''' Get the singleton instance of the class defined for the loader. '''
        if not self._instance:
            self._instance = self.fork()

        return self._instance

    def fork(self):
        '''
        :Version: 1.5

        Fork an instance of the class defined for the loader.
        '''
        self.__prepare()

        cls = self._loader.package

        try:
            instance = cls(*self._args, **self._kwargs)
        except TypeError as e:
            signature = inspect.signature(cls)

            given_args = [str(a) for a in self._args]

            for k in signature.parameters:
                if k not in self._kwargs:
                    continue

                value = self._kwargs[k]
                kind  = type(value).__name__ or 'null'

                given_args.append(
                    '{key}={value} : {kind}'.format(
                        key = k,
                        kind = kind,
                        value = '"{}"'.format(value) \
                            if isinstance(value, str) \
                            else value
                    )
                )

            error_message = ' '.join([
                'The configuration for the entity "{entity_id}" of class',
                '{full_class_name}{signature} is POSSIBLY INCORRECT, given',
                'that the provided configuration shows that you provide ({given}).'
            ]).format(
                entity_id       = self._id,
                full_class_name = self._loader._path,
                signature       = signature,
                given           = ', '.join(given_args),
            )

            raise InstantiationError(error_message)

        # Return the instance if this entity is not interceptable.
        if not self.interceptable:
            return instance

        # For each PUBLIC method, make it interceptable with Action.
        self._bind_interceptions(instance, self._interceptions)

        return instance

    def __prepare(self):
        if self._prepared or self._instance:
            return

        for i in range(len(self._args)):
            if isinstance(self._args[i], Proxy):
                self._args[i] = self._args[i].load()

        for i in self._kwargs:
            if isinstance(self._kwargs[i], Proxy):
                self._kwargs[i] = self._kwargs[i].load()

        self._prepared = True



class DeprecatedAPI(Exception):
    ''' Exception thrown when a deprecated API is used. '''

class MisplacedValidatorError(Exception):
    ''' Exception thrown when a validator is used with type (e.g., class, perimitive types etc.) '''

class ForbiddenForkError(Exception):
    ''' Exception thrown only when a locator try to fork a proxy. '''

class DuplicateKeyWarning(Warning):
    ''' Warning thrown when an internal dictionary detects an attempt of re-assignment to the direction by the existed key. '''

class MultipleInterceptingEventsWarning(Warning):
    ''' Warning thrown when more than one intercepting event is detected. '''

class IncompatibleBlockError(Exception):
    ''' Exception thrown when :meth:`imagination.locator.Locator.load_xml` cannot process the entity block. '''

class UnknownFileError(Exception):
    ''' Exception thrown when :meth:`imagination.locator.Locator.load_xml` cannot locate the file on the given path. '''

class UnknownEntityError(Exception):
    ''' Exception thrown when :class:`imagination.locator.Locator` constructor receives an unusable entity. '''

class UnknownProxyError(Exception):
    ''' Exception thrown when :class:`imagination.helper.assembler.Assembler` could not find the proxy. '''

class NonCallableError(Exception):
    ''' Exception thrown when :class:`imagination.action.Action` tries to execute a non-callable reference. '''

class LockedEntityException(Exception):
    '''
    Exception thrown when a caller attempts to update any of alterable properties
    of an instance of :class:`imagination.entity.Entity` when it is in the locked
    state.
    '''

class InstantiationError(RuntimeError):
    """ Exception thrown when forking fails. """



from re import split

from kotoba.kotoba import Kotoba
from kotoba        import load_from_file

from imagination.decorator.validator import restrict_type
from imagination.entity              import *
from imagination.factorization       import Factorization, NotReadyError
from imagination.exception           import *
from imagination.loader              import Loader
from imagination.helper.data         import *
from imagination.meta.aspect         import Contact
from imagination.meta.interception   import Interception
from imagination.meta.package        import Parameter
from imagination.proxy               import Proxy

from . import extension
from . import mixin

class Assembler(mixin.ParameterParsingMixin):
    """
    The entity assembler via configuration files.

    :param imagination.helper.data.Transformer transformer: the data transformer

    .. versionadded:: 1.5

        This class is added.

    .. versionchanged:: 1.9

        Added the support for entity factorization.

    .. versionchanged:: 1.10

        Added the support for delayed injections when the factory service is not ready to use during the factorization.

    .. versionchanged:: 1.20

        Supports callable proxies, execute-only-once proxies, and reference proxies.

    """
    __known_interceptable_events = ['before', 'pre', 'post', 'after']
    __entity_element_required_attributes = {
        'entity': {
            'id':    'entity identifier',
            'class': 'entity class',
        },
        'factorization': {
            'id':   'factorized entity identifier',
            'with': 'factory entity identifier',
            'call': 'factory method',
        },
        'callable': {
            'id':     'execution identifier',
            'method': 'callable method path (used by the "import" statement)',
            'cached': 'whether to cache the result'
        },
    }

    @restrict_type(Transformer)
    def __init__(self, transformer):
        self._transformer = transformer

        self.__interceptions = []
        self.__known_proxies = {}
        self.__delay_injection_map = {}

        self.__extensions = [
            extension.EntityRegistrar(self._transformer),
            extension.FactorizationRegistrar(self._transformer),
            extension.CallableRegistrar(self._transformer),
        ]

    def activate_passive_loading(self):
        """ Activate the passive loading mode.

            This is to delay the instantiation or forking of an entity to when
            it is referred/queried but undefined. This assumes that the queried
            entity will later be defined.

            If later on it isn't defined, the exception will be thrown.

            When this method is used, please do not forget to call :meth:`deactivate_passive_loading`
            to disable the passive mode.

            .. versionadded:: 1.7
        """
        self.locator.in_passive_mode = True

    def deactivate_passive_loading(self):
        """ Deactivate the passive loading mode.

            When :meth:`activate_passive_loading` is used, this method must be
            called to disable the passive mode.

            .. versionadded:: 1.7
        """

        if self.__delay_injection_map:
            injection_order = []

            for k in self.__delay_injection_map:
                di = self.__delay_injection_map[k]
                n  = di['node']
                p  = di['ping']
                r  = di['ready']

                if not n or r: # Skip if either the node is not defined or the delay injection is done for this entry.
                    continue

                di['ready'] = True

                injection_order.append((k, di['node'], di['ping']))

            injection_order.sort(key=lambda i: i[2])

            for di in injection_order:
                id, node, ping = di

                self.__get_interceptions(node)
                self.__register_entity(node)

            self.__activate_interceptions()

        self.locator.in_passive_mode = False

    def load(self, filepath):
        """
        Load the configuration.

        :param str filepath: the file path to the configuration.
        """
        xml = load_from_file(filepath)

        # First, register proxies to entities (for lazy initialization).
        for node in xml.children():
            self.__validate_node(node)
            self.__register_proxy(node)

        # Then, register loaders for entities.
        for node in xml.children():
            self.__get_interceptions(node)
            self.__register_entity(node)

        self.__activate_interceptions()

    def __activate_interceptions(self):
        # Then, declare interceptions to target entities.
        for interception in self.__interceptions:
            self.locator\
                .get_wrapper(interception.actor.id)\
                .register_interception(interception)

        self.__interceptions = []

    @property
    def locator(self):
        """
        The injected locator via the data transformer.

        :rtype: imagination.locator.Locator
        """
        return self._transformer.locator()

    @restrict_type(Kotoba)
    def __validate_node(self, node):
        """ .. deprecated:: 1.20 """
        entity_type = node.name()

        for ext in self.__extensions:
            if entity_type not in ext.element_names():
                continue

            ext.validate(node)

    @restrict_type(Kotoba)
    def __register_proxy(self, node):
        id    = node.attribute('id')
        proxy = Proxy(self.locator, id)

        self.locator.set(id, proxy)

        # this is for interceptors
        self.__known_proxies[id] = proxy

    @restrict_type(Kotoba)
    def __register_entity(self, node):
        entity_type = node.name()

        try:
            for ext in self.__extensions:
                if entity_type not in ext.element_names():
                    continue

                ext.register(node)

        except NotReadyError as e:
            required = str(e)
            targeted = node.attribute('id')

            self.__set_delay_injection(required, None, True)
            self.__set_delay_injection(targeted, node, False)

    def __set_delay_injection(self, id, node, ping):
        if id not in self.__delay_injection_map:
            self.__delay_injection_map[id] = {
                'ping':  0,
                'node':  node,
                'ready': False,
            }

        injection = self.__delay_injection_map[id]

        if node:
            injection['node'] = node

        if ping:
            injection['ping'] -= 1

    @restrict_type(Kotoba)
    def __get_interceptions(self, node):
        for sub_node in node.children('interception'):
            self.__interceptions.append(self.__get_interception(sub_node))

    @restrict_type(Kotoba)
    def __get_interception(self, node):
        actor = None
        event = None

        intercepted_action = None
        handling_action    = None

        for given_event in self.__known_interceptable_events:
            given_actor = node.attribute(given_event)

            # If the actor is not defined, continue or if the event is already
            # set (in the earlier iteration), raise the exception.
            if not given_actor:
                continue
            elif event:
                raise MultipleInterceptingEventsWarning(given_event)

            # Initially get the name of the actor and the handler.
            actor   = given_actor
            handler = node.parent().attribute('id')

            if actor == Interception.self_reference_keyword():
                actor = handler

            # If the actor or the handler has no proxies, raise the exception.
            if actor not in self.__known_proxies:
                raise UnknownProxyError('The target ({}) of the interception is unknown.'.format(actor))

            if handler not in self.__known_proxies:
                raise UnknownProxyError('The handler ({}) of the interception is unknown.'.format(handler))

            actor   = Contact(self.__known_proxies[actor], node.attribute('do'))
            handler = Contact(self.__known_proxies[handler], node.attribute('with'), self._get_params(node))
            event   = given_event

        return Interception(event, actor, handler)



try:
    from setuptools import setup
except:
    from distutils.core import setup

setup(
    name         = 'imagination',
    version      = '1.29.0',
    description  = 'Reusable Component Framework',
    author       = 'Juti Noppornpitak',
    author_email = 'juti_n@yahoo.co.jp',
    url          = 'https://github.com/shiroyuki/Imagination',
    packages     = ['imagination', 'imagination.decorator', 'imagination.helper', 'imagination.meta']
)



from imagination.action              import Action
from imagination.decorator.validator import restrict_type

class InterceptableObject(object):
    def __init__(self):
        self._locked        = False
        self._interceptable = False
        self._interceptions = []

    def lock(self):
        """ Lock the object """
        self._locked = True

    def unlock(self):
        """ Unlock the object """
        self._locked = True

    @property
    def locked(self):
        """ Flag indicating whether the object is locked

            :rtype: bool
        """
        return self._locked

    @property
    def interceptable(self):
        '''
        Flag if this entity is interceptable

        :rtype: boolean
        '''
        return self._interceptable

    @interceptable.setter
    @restrict_type(bool)
    def interceptable(self, interceptable):
        '''
        Define if this entity is interceptable.

        :param interceptable: Flag if this entity is interceptable
        :type  interceptable: boolean
        '''
        if self.locked:
            raise LockedEntityException

        self._interceptable = interceptable

    @property
    def interceptions(self):
        '''
        Retrieve the list of interceptions.
        '''
        return self._interceptions

    @interceptions.setter
    @restrict_type(list)
    def interceptions(self, interceptions):
        '''
        Define the list of interceptions.

        :param interceptions: list of interceptions
        :type  interceptions: list
        '''
        self._interceptions = interceptions

    def register_interception(self, interception):
        self._interceptions.append(interception)

    def _bind_interceptions(self, instance, interceptions):
        for attribute in dir(instance):
            if attribute[0] == '_':
                continue

            ref = instance.__getattribute__(attribute)

            if not callable(ref) or isinstance(ref, Action):
                continue

            new_ref = Action(ref)

            for interception in interceptions:
                if interception.actor.method_name != attribute:
                    continue

                new_ref.register(interception)

            try:
                instance.__setattr__(attribute, new_ref)
            except AttributeError:
                pass



from kotoba.kotoba import Kotoba

from imagination.entity        import *
from imagination.factorization import *
from imagination.exception     import *
from imagination.helper        import mixin

class AbstractRegistrar(mixin.ParameterParsingMixin):
    def __init__(self, transformer):
        self._transformer = transformer

    @property
    def locator(self):
        """
        The injected locator via the data transformer.

        :rtype: imagination.locator.Locator
        """
        return self._transformer.locator()

    def required_attributes(self):
        raise NotImplementedError('The attribute-name-to-description map must be defined.')

    @restrict_type(Kotoba)
    def validate(self, node):
        entity_type = node.name()

        required_attributes = self.required_attributes()

        for name in required_attributes:
            if node.attribute(name):
                continue

            raise IncompatibleBlockError(
                'Block "{}" needs the "{}" attribute being defined.'.format(
                    entity_type, required_attributes[name]
                )
            )

    def register(self, node):
        entity_id = node.attribute('id')

        params = self._get_params(node)
        entity = self.construct(node, params)

        entity.interceptable = self._transformer.cast(node.attribute('interceptable') or 'true', 'bool')
        entity.tags          = self.__get_tags(node)

        self.locator.set(entity_id, entity)

    def construct(self, node, params):
        raise NotImplementedError('Required to implement')

    @restrict_type(Kotoba)
    def __get_tags(self, node):
        tags = node.attribute('tags')

        return tags and split(' ', tags.strip()) or []

class EntityRegistrar(AbstractRegistrar):
    def element_names(self):
        return ['entity']

    def required_attributes(self):
        return {
            'id':    'entity identifier',
            'class': 'entity class',
        }

    def construct(self, node, params):
        id     = node.attribute('id')
        kind   = node.attribute('class')
        loader = Loader(kind)

        return Entity(id, loader, *params.largs, **params.kwargs)

class FactorizationRegistrar(AbstractRegistrar):
    def element_names(self):
        return ['factorization']

    def required_attributes(self):
        return {
            'id':   'factorized entity identifier',
            'with': 'factory entity identifier',
            'call': 'factory method',
        }

    def construct(self, node, params):
        factory_id     = node.attribute('with')
        factory_method = node.attribute('call')

        return Factorization(self.locator, factory_id, factory_method, params)

class CallableRegistrar(AbstractRegistrar):
    def element_names(self):
        return ['callable']

    def required_attributes(self):
        return {
            'id':     'callable identifier',
            'method': 'callable method path (used by the "import" statement)',
            'static': 'the flag to cache the end result'
        }

    def construct(self, node, params):
        entity_type = node.name()
        method_path = node.attribute('method')

        is_static = self._transformer.cast(node.attribute('static') or 'true', 'bool')

        ProxyClass = OnceProxy if entity_type == 'definition' else CallbackProxy

        return ProxyClass(method_path, params.largs, params.kwargs, is_static)
